# $Id: jeff_complex_cmp.py,v 1.4 2012-10-03 01:19:47 jdeifik Exp $
# Copyright Jeff turbo Deifik 1997
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# These functions are used by uniq and sim to process and compare lines.
# This is similar to compare.c, except it has some context. This
# simplifies the interface, and is faster besides.
# Written by Jeff turbo Deifik @1997 (c), Dec-11-2008 (python)


from __future__ import print_function

import jefflib		# Pop_Count, Print_Error, Editdist


class jeff_complex_cmp:


    # Create the Jeff_complex_compare class
    # This is where the real initialization takes place.
    def __init__(self, Flags, M_num):
        # validate input
        pop = jefflib.Pop_Count(['CMP_PERCENT' in Flags,
                                 'CMP_ABSNUM' in Flags,
                                 'CMP_MISMATCH' in Flags])
        if pop > 1:
            jefflib.Print_Error("Too many flags specified")
        if pop < 1:
            jefflib.Print_Error("No percent, absnum, or mismatch specified")

        self.__ed          = jefflib.Editdist()
        self.__flags       = Flags
        self.__m_num       = M_num
        self.__line        = ''
        self.__len         = 0

# The rest of the variables are strictly for optimization. The idea is that
# these computations can be done once, at init time, rather than n**2 times.
        self.__flags_and_complex     = 'CMP_COMPLEX' in Flags
        self.__flags_and_local       = 'CMP_LOCAL' in Flags
        self.__flags_and_percent     = 'CMP_PERCENT' in Flags
        self.__flags_and_first       = 'CMP_FIRST' in Flags
        self.__flags_and_absnum      = 'CMP_ABSNUM' in Flags
        self.__flags_and_mismatch    = 'CMP_MISMATCH' in Flags
        self.__flags_and_per_and_not_first = 'CMP_PERCENT' in Flags and \
            not ('CMP_FIRST' in Flags)


    # This is one line that we will be comparing against. Since we will be
    # doing many comparisons with this line, we need only pass it in once.
    def Set_Line(self, line1):
        self.__line = line1
        self.__len  = len(line1)


    # Return something non-negative if line1 equals line2, else return -1.
    def Global(self, line2, len2):
        # Here is the gory processing for the m flag.
        if self.__flags_and_complex:
                        # Find the max and min line lengths.
            max_len = max(self.__len, len2)
            min_len = min(self.__len, len2)

            # Find the matching number, based on m_num, CMP_PERCENT, CMP_ABSNUM,
            # CMP_MISMATCH and CMP_FIRST.
            m = self.__compute_m(max_len)
            if self.__check_percent_ok(max_len, min_len) == False: return(-1)

            # If a string is shorter than the number of characters they must
            # match, fail.
            if m > min_len: return(-1)
            if m == 0: return(1)
# I question if it is possible to execute __compar_local from here.
# You would need to call Global, but have local set...
            if self.__flags_and_local:		# Do a local comparison.
                return(self.__compar_local(line2, len2, m))
            else:				# Do a global comparison.
                return(self.__compar_global(line2, len2, max_len, m))
        else:					# Do a simple minded comparison.
            if self.__len != len2: return(-1)
            if self.__line == line2: return(1)
            else: return(-1)


    # Return 1 if line1 equals line2, else return -1.
    # If special changes, then re-do all the context for BM_Search.
    def Local(self, line2, len2, special):
        # Here is the gory processing for the m flag.
        if self.__flags_and_complex:
                        # Find the max and min line lengths.
            max_len = max(self.__len, len2)
            min_len = min(self.__len, len2)

            # Find the matching number, based on m_num, CMP_PERCENT, CMP_ABSNUM
            # CMP_MISMATCH and CMP_FIRST.
            m = self.__compute_m(max_len)
            if self.__check_percent_ok(max_len, min_len) == False: return(-1)

            # If a string is shorter than the number of characters they must
            # match, fail.
            if m > min_len: return(-1)
            if m == 0: return(1)
            if self.__flags_and_local:		# Do a local comparison.
                if self.__check_local_ok(len2, m) == False: return(-1)
# Now we call BM_Search for all substrings of length m in string line1.
                for i in range(0, self.__len - m + 1):
                    s = self.__line[i:i+m]	# This is expensive
                    res =  line2.find(s)	# So is this
                    if res != -1: return(1)
                return(-1)
            else:				# Do a global comparison.
                return(self.__compar_global(line2, max_len, m))
        else:			# Do a simple minded comparison.
            if self.__len != len2: return(-1)
            if self.__line == line2: return(1)
            else: return(-1)


    def __check_percent_ok(self, max_len, min_len):
        # If we are using percent match, and not min.
        if self.__flags_and_per_and_not_first:
            # See if the percentage difference of the length of string1 and
            # string2 exceeds the allowable percentage difference, if so fail.
            # If the longest string is of length 0 comparison is vacously true.
            if max_len == 0: return(True)

            tmp_min = min_len * 100
            tmp_max = self.__m_num * max_len
            return(tmp_min >= tmp_max)
        return(True)


    def __check_local_ok(self, len2, m):
        if self.__len < m or len2 < m: return(False)
        return(True)


    # Compute the number of matching characters required in order for
    # the match to be considered successful. +
    def __compute_m(self, max_len):
        if self.__flags_and_first: mult = self.__len
        else: mult = max_len

        if self.__flags_and_percent:
            # The integer way. Faster in general, due to no int-fp and
            # fp-int conversions.
            return(mult * self.__m_num + 50)//100

        if self.__flags_and_absnum:
            return(self.__m_num)
        if self.__flags_and_mismatch:
            return (mult - self.__m_num)


    # Do a global comparison, returning j, the edit distance (non-negative)
    # if line1 equals line2, else return -1.
    def __compar_global(self, line2, len2, max_len, m):
        ed = jefflib.Editdist()	# Init ed to be an editdist object
        j = ed.UKKA_Dist_With_Max(self.__line, self.__len, line2, len2,
                                  1 + max_len - m)
        k = max_len - j		# Maxlen - distance.
        if k >=m: return(j)
        else: return -1


    # Do a local comparison, returning 1 if line1 equals line2, else return -1.
    def __compar_local(self, line2, len2, m):
        if self.__check_local_ok(len2, m) == False: return(-1)

        for i in range(0, self.__len - m + 1):
            s = self.__line[i:i+m]			# This is expensive
            for j in range(0, len2 - m + 1):
                s2 = line2[j:j+m]			# This is expensive
                assert(0)
                if s == s2: return 1
        return(-1)
